﻿using System;
using System.Collections.Generic;
using System.ComponentModel;
using System.Data;
using System.Drawing;
using System.Linq;
using System.Text;
using System.Windows.Forms;
using System.Threading.Tasks;

namespace OptimisationCodeCarteBundling
{

    /// <summary>
    /// You can Find in the code below 2 versions of bundling algorithm . the first use Grad
    /// Gradient Kernel Splatting and the other use kernel values splatting. 
    /// to use the other algorithm , a code in the timer ticks  in the region Density Map
    /// should be un-Comment and comment the region gradient below.
    /// </summary>
    public partial class FormBundling : Form
    {
        Pen TheDrawingPen = new Pen(Color.FromArgb(50, 0, 0, 0));
        SolidBrush TheDrawingBrush = new SolidBrush(Color.FromArgb(100, 255, 0, 0));

        public FormBundling()
        {
            InitializeComponent();
        }

        #region Parameter Initializing

        #region Gradient declaration
        /// <summary>
        /// Are used to cumulate the gradient. They play the role of density Map
        /// </summary>
        public float[,] GradientDensityMapX;
        public float[,] GradientDensityMapY;

        /// <summary>
        /// store the X and Y  kernel gradient values 
        /// </summary>
        public float[] GradientKernelX;
        public float[] GradientKernelY;

        /// <summary>
        ///  store min and max of the gradient cumulative values for X coordinate
        /// </summary>
        float minGradientDensValueX;
        float maxGradientDensValueX;

        /// <summary>
        /// store min and max of the gradient cumulative values for Y coordinate
        /// </summary>
        float minGradientDensValueY;
        float maxGradientDensValueY;

        #endregion


        #region Density Map declaration

        /// <summary>
        /// Initialize the  densityMap
        /// </summary>
        public int DensityMapSize = 500;
        public float[,] Densitymap;

        float minDensValue;
        float maxDensValue;

        /// <summary>
        /// initialize the bitmap
        /// </summary>
        public Bitmap Bmp;

        private VectorGrad[,] GradCrt;

        #endregion


        #region Kernel Components

        /// <summary>
        /// 1 dimension array of kernel
        /// </summary>
        public int KernelSize = 25;
        public float[] Kernel;

        #endregion


        /// <summary>
        /// to give the permission to store points in the list while mouse moving
        /// </summary>
        bool drawPointsFlag = false;
        Graphics g;

        #region Mouse Interaction declaration

        /// <summary>
        /// lists to stock points when user click on the right click 
        /// </summary>
        List<PointF> mouseDrawpointsList;
        List<PointF[]> mouseDrawPoints = new List<PointF[]>();
        PointF[][] DisplayDrawPoints;

        // Attributs for left mouse points 
        List<PointF[]> mouseDrawparticules = new List<PointF[]>();

        #endregion

        /// <summary>
        /// to give the user a choice to choose between show points or not
        /// </summary>
        bool IsShowVertices = false;

        /// <summary>
        /// it will set the attraction force to bundle between lines
        /// </summary>
        public float AttractionFactor = 2.1f;

        /// <summary>
        /// smoothing limit to set the smoothing degree (1 = joining 3 nodes together)
        /// </summary>
        int SmoothingLimit = 1;

        /// <summary>
        /// the distance where we should split the distance and add a node if two nodes are too far or remove it if too close
        /// </summary>
        double splitDist = 15;
        double removeDist = 5;

        #endregion

        private void Form1_Load(object sender, EventArgs e)
        {

            #region Initializing components

            Densitymap = new float[DensityMapSize, DensityMapSize];
            Bmp = new Bitmap(DensityMapSize, DensityMapSize);
            Kernel = new float[KernelSize * KernelSize];

            trackBarAttractionFactor.Value = Convert.ToInt32(AttractionFactor);
            numericUpDownKernelSize.Value = KernelSize;
            labelSmoothingFactor.Text = SmoothingLimit.ToString();
            labelAttractionFactor.Text = AttractionFactor.ToString();

            #endregion


            Kernel = ComputeKernelValues(KernelSize);
            ComputeKernelGrad(Kernel, out GradientKernelX, out GradientKernelY);

            GradientDensityMapX = new float[DensityMapSize, DensityMapSize];
            GradientDensityMapY = new float[DensityMapSize, DensityMapSize];

        }


        private void pictureBox1_Paint(object sender, PaintEventArgs e)
        {
            g = e.Graphics;
            g.SmoothingMode = System.Drawing.Drawing2D.SmoothingMode.AntiAlias;

            if (DisplayDrawPoints == null) return;

            #region Right Click
            //draw line
            foreach (var line in DisplayDrawPoints)
            {
                g.DrawLines(TheDrawingPen, line);
            }
            /// draw points on screen if checkbox is ticked
            if (IsShowVertices)
            {
                foreach (var line in DisplayDrawPoints)
                {
                    foreach (var pt in line)
                    {
                        g.FillEllipse(TheDrawingBrush, pt.X - 2, pt.Y - 2, 4, 4);
                    }
                }
            }
            #endregion

        }

        /// <summary>
        /// it will compute distance between nodes and add one if are far enough or remove one if are too close
        /// </summary>
        /// <param name="splitDist"></param>
        /// <param name="removeDist"></param>
        /// <param name="targetArray"></param>
        private void Resampling(double splitDist, double removeDist, PointF[][] targetArray)
        {
            PointF currentVert = new PointF();
            PointF nextVert = new PointF();
            List<PointF> tmpVertexList;

            if (targetArray.Length < 3) { return; }
            for (int x = 0; x < targetArray.Length; x++)
            {
                tmpVertexList = new List<PointF>();
                tmpVertexList.Add(targetArray[x][0]);

                // Sampling on each lines (= trajectory)

                for (int y = 0; y < targetArray[x].Length - 1; y++)
                {
                    currentVert = targetArray[x][y];
                    nextVert = targetArray[x][y + 1];
                    double dist = Math.Sqrt((currentVert.X - nextVert.X) * (currentVert.X - nextVert.X) + (currentVert.Y - nextVert.Y) * (currentVert.Y - nextVert.Y));

                    // Test if the next point is too far or too close
                    if (dist > splitDist)
                    {
                        tmpVertexList.Add(new Point((int)((currentVert.X + nextVert.X) / 2.0), (int)((currentVert.Y + nextVert.Y) / 2.0)));
                    }
                    if (!(dist < removeDist) || y == targetArray.Length - 2)
                    {
                        tmpVertexList.Add(nextVert);
                    }
                }
                tmpVertexList.Add(targetArray[x][targetArray[x].Length - 1]);
                targetArray[x] = tmpVertexList.ToArray();
            }

        }


        /// <summary>
        /// Smooth the lines with a lalacien filter
        /// </summary>
        /// <param name="targetArray"></param>
        private void smoothing(PointF[][] targetArray)
        {
            int smoothnbr = 0;
            PointF tempf = new PointF(); ;
            if (SmoothingLimit != 0)
            {
                foreach (var line in targetArray)
                {

                    for (int i = 1; i < line.Length - 1; i++)
                    {
                        smoothnbr = 0;
                        tempf = new PointF(); ;

                        for (int t = -SmoothingLimit; t <= SmoothingLimit; ++t)
                        {
                            if (i + t < 0 || i + t > line.Length - 1) { continue; }
                            tempf.X += line[i + t].X;
                            tempf.Y += line[i + t].Y;
                            smoothnbr++;
                        }
                        tempf.X /= smoothnbr;
                        tempf.Y /= smoothnbr;

                        line[i].X = tempf.X;
                        line[i].Y = tempf.Y;

                    }
                }
            }
        }

        private void timerUpdate_Tick(object sender, EventArgs e)
        {
            if (DisplayDrawPoints == null)
                return;


            //debug get the total number of points
            int totalpoints = 0;
            Resampling(splitDist, removeDist, DisplayDrawPoints);


            #region Density Map region
            //ComputeDensityMap(DisplayDrawPoints);
            //ComputeGrad();
            //   UpdateLine();
            #endregion

            #region Kernel Gradient
            ComputeDensityMapKernelGradient(DisplayDrawPoints);
            NormalizeGradientMap();
            UpdateLineGradKernel();

            #endregion

            smoothing(DisplayDrawPoints);
            pictureBox1.Invalidate();

            // it shows total number of points
            foreach (var line in DisplayDrawPoints)
            {
                totalpoints += line.Length;
            }
            labelNumberOfPoints.Text = totalpoints.ToString();

        }

        /// <summary>
        /// Nomalize the gradient map
        /// </summary>
        private void NormalizeGradientMap()
        {
            int nbrItems = GradientDensityMapX.Length;

            // float min = Math.Min(minGradientDensValueX, minGradientDensValueY);
            //   float max = Math.Max(maxGradientDensValueX, maxGradientDensValueY);

            Parallel.For(0, DensityMapSize, y =>
            {
                float l = 0;
                for (int x = 0; x < DensityMapSize; x++)
                {
                    l = (float)Math.Sqrt(GradientDensityMapX[x, y] * GradientDensityMapX[x, y] + GradientDensityMapY[x, y] * GradientDensityMapY[x, y]);
                    if (l != 0)
                    {
                        GradientDensityMapX[x, y] /= l;
                        GradientDensityMapY[x, y] /= l;
                    }

                }
            });

        }

        #region Density Map Component

        private void GetMaxMinFromDensityMap()
        {
            maxDensValue = float.MinValue;
            minDensValue = float.MaxValue;

            for (int y = 0; y < DensityMapSize; y++)
            {
                for (int x = 0; x < DensityMapSize; x++)
                {
                    if (Densitymap[x, y] > maxDensValue)
                        maxDensValue = Densitymap[x, y];
                    if (Densitymap[x, y] < minDensValue)
                        minDensValue = Densitymap[x, y];
                }
            }
        }

        private void ComputeBitmapFromDensityMap()
        {

            for (int y = 0; y < DensityMapSize; y++)
            {
                for (int x = 0; x < DensityMapSize; x++)
                {
                    Bmp.SetPixel(x, y, Color.Black);
                }
            }
        }


        private void ComputeDensityMap(PointF[][] targetArray)
        {
            SetDensityMapToZero();

            foreach (var line in targetArray)
            {
                foreach (var point in line)
                {
                    PointF points = new PointF(point.X - KernelSize / 2, point.Y - KernelSize / 2);
                    for (int y = 0; y < KernelSize; y++)
                    {
                        for (int x = 0; x < KernelSize; x++)
                        {
                            if (((points.X + x) < DensityMapSize) && ((points.Y + y) < DensityMapSize) && ((points.X + x) > 0) && ((points.Y + y) > 0))
                            {
                                Densitymap[Convert.ToInt32(points.X + x), Convert.ToInt32(points.Y + y)] += Kernel[x + y * KernelSize];
                            }
                        }
                    }

                }
            }
            GetMaxMinFromDensityMap();
        }

        private void UpdateLine()
        {
            if (DisplayDrawPoints == null)
                return;

            foreach (var line in DisplayDrawPoints)
            {
                for (int i = 1; i < line.Length - 1; i++)
                {
                    PointF item = line[i];
                    line[i].X += GradCrt[Convert.ToInt32(item.X), Convert.ToInt32(item.Y)].X * AttractionFactor;
                    line[i].Y += GradCrt[Convert.ToInt32(item.X), Convert.ToInt32(item.Y)].Y * AttractionFactor;
                }
            }

        }

        public void ComputeGrad()
        {
            GradCrt = new VectorGrad[DensityMapSize, DensityMapSize];

            for (int y = 0; y < DensityMapSize; y++)
            {
                for (int x = 0; x < DensityMapSize; x++)
                {
                    GradCrt[x, y] = new VectorGrad();
                    if ((x < (DensityMapSize - 1)) && (y < (DensityMapSize - 1)))
                    {
                        GradCrt[x, y].X = Densitymap[x + 1, y] - Densitymap[x, y];
                        GradCrt[x, y].Y = Densitymap[x, y + 1] - Densitymap[x, y];
                        GradCrt[x, y].Normalize();
                    }
                }
            }
        }


        private void SetDensityMapToZero()
        {
            for (int y = 0; y < DensityMapSize; y++)
            {
                for (int x = 0; x < DensityMapSize; x++)
                {
                    Densitymap[x, y] = 0;
                }
            }

        }

        #endregion

        #region Gradient Components


        #region Compute and store gradient vectors of the kernel

        public void ComputeKernelGrad(float[] kernel, out float[] gradientKernelX, out float[] gradientKernelY)
        {
            gradientKernelX = new float[KernelSize * KernelSize];
            gradientKernelY = new float[KernelSize * KernelSize];

            for (int y = 0; y < (KernelSize); y++)
            {

                for (int x = 0; x < (KernelSize); x++)
                {
                    if ((x > 0) && (x < (KernelSize - 1)))
                        gradientKernelX[x + y * KernelSize] = kernel[(x - 1) + y * KernelSize] - kernel[(x + 1) + y * KernelSize];
                    if (x == 0)
                        gradientKernelX[x + y * KernelSize] = kernel[(x + 1) + y * KernelSize] - kernel[(x) + y * KernelSize];
                    if (x == (KernelSize - 1))
                        gradientKernelX[x + y * KernelSize] = kernel[(x) + y * KernelSize] - kernel[(x - 1) + y * KernelSize];

                    if ((y > 0) && (y < (KernelSize - 1)))
                        gradientKernelY[x + y * KernelSize] = kernel[(x) + (y - 1) * KernelSize] - kernel[(x) + (y + 1) * KernelSize];
                    if (y == 0)
                        gradientKernelY[x + y * KernelSize] = kernel[(x) + (y + 1) * KernelSize] - kernel[(x) + (y) * KernelSize];
                    if (y == (KernelSize - 1))
                        gradientKernelY[x + y * KernelSize] = kernel[(x) + (y) * KernelSize] - kernel[(x) + (y - 1) * KernelSize];

                }
            }

            //Check that the sum is zero
            //float sumX = 0;
            //foreach (var item in gradientKernelX)
            //{
            //    sumX += item;
            //}
            //float sumY = 0;
            //foreach (var item in gradientKernelY)
            //{
            //    sumY += item;
            //}
        }

        #endregion


        /// <summary>
        /// it will add the Gradients values of X and Y in two array X and Y 
        /// </summary>
        /// <param name="targetArray"></param>
        private void ComputeDensityMapKernelGradient(PointF[][] targetArray)
        {
            SetGradientDensityMapToZero();


            foreach (var line in targetArray)
            {
                foreach (var point in line)
                {
                    PointF points = new PointF(point.X - KernelSize / 2, point.Y - KernelSize / 2);
                    for (int y = 0; y < KernelSize; y++)
                    {
                        for (int x = 0; x < KernelSize; x++)
                        {
                            if (((points.X + x) < DensityMapSize) && ((points.Y + y) < DensityMapSize) && ((points.X + x) > 0) && ((points.Y + y) > 0))
                            {
                                GradientDensityMapX[Convert.ToInt32(points.X + x), Convert.ToInt32(points.Y + y)] += GradientKernelX[x + y * KernelSize];
                                GradientDensityMapY[Convert.ToInt32(points.X + x), Convert.ToInt32(points.Y + y)] += GradientKernelY[x + y * KernelSize];
                            }
                        }
                    }

                }
            }
            GetMaxMinFromGradientDensityMap();
        }

        private void SetGradientDensityMapToZero()
        {
            for (int y = 0; y < DensityMapSize; y++)
            {
                for (int x = 0; x < DensityMapSize; x++)
                {
                    GradientDensityMapX[x, y] = 0;
                    GradientDensityMapY[x, y] = 0;
                }
            }
        }

        private void UpdateLineGradKernel()
        {
            if (DisplayDrawPoints == null)
                return;

            foreach (var line in DisplayDrawPoints)
            {
                for (int i = 1; i < line.Length - 1; i++)
                {
                    if (((line[i].X) < DensityMapSize) && ((line[i].Y) < DensityMapSize) && ((line[i].X) > 0) && ((line[i].Y) > 0))
                    {
                        PointF temp = line[i];
                        line[i].X -= GradientDensityMapX[Convert.ToInt32(temp.X), Convert.ToInt32(temp.Y)] * AttractionFactor;
                        line[i].Y -= GradientDensityMapY[Convert.ToInt32(temp.X), Convert.ToInt32(temp.Y)] * AttractionFactor;
                    }

                }
            }

        }


        private void GetMaxMinFromGradientDensityMap()
        {
            maxGradientDensValueX = float.MinValue;
            minGradientDensValueX = float.MaxValue;
            maxGradientDensValueY = float.MinValue;
            minGradientDensValueY = float.MaxValue;

            for (int y = 0; y < DensityMapSize; y++)
            {
                for (int x = 0; x < DensityMapSize; x++)
                {
                    if (GradientDensityMapX[x, y] > maxGradientDensValueX)
                        maxGradientDensValueX = GradientDensityMapX[x, y];
                    if (GradientDensityMapX[x, y] < minGradientDensValueX)
                        minGradientDensValueX = GradientDensityMapX[x, y];

                    if (GradientDensityMapY[x, y] > maxGradientDensValueY)
                        maxGradientDensValueY = GradientDensityMapY[x, y];
                    if (GradientDensityMapY[x, y] < minGradientDensValueY)
                        minGradientDensValueY = GradientDensityMapY[x, y];
                }
            }
        }

        #endregion

        #region Compute kernel points Distance

        #region Compute kernel Values (distance between each square from the center )

        private float[] ComputeKernelValues(float KernelSizeP)
        {
            float[] kernel = new float[(int)(KernelSizeP * KernelSizeP)];

            float halfKernel = (float)((KernelSizeP / 2.0) * (KernelSizeP / 2.0));

            for (float y = 0; y < KernelSizeP; y++)
            {
                for (float x = 0; x < KernelSizeP; x++)
                {
                    float centerDist = (float)(((float)(y - (KernelSizeP - 1.0) / 2.0)) * ((float)(y - (KernelSizeP - 1.0) / 2.0)) +
                        ((x - (KernelSizeP - 1.0) / 2.0)) * ((x - (KernelSizeP - 1.0) / 2.0)));

                    if (centerDist < halfKernel)
                    {
                        double scaledValue = Scale(centerDist, 0, halfKernel, 1, 0);
                        kernel[(int)(x + y * KernelSizeP)] = (float)scaledValue;
                    }

                }
            }
            return kernel;
        }

        #endregion

        #region Scale Function get 0 if distance is >> than kernelsize/2 and 1 if it's equal.
        private double Scale(double value, double min, double max, double minScale, double maxScale)
        {
            double scaled = minScale + (double)(value - min) / (max - min) * (maxScale - minScale);
            return scaled;
        }
        #endregion

        #endregion

        #region Mouse Interaction

        /// <summary>
        /// when left mouse click it will create list to stock points
        /// </summary>
        /// <param name="sender"></param>
        /// <param name="e"></param>
        private void pictureBox1_MouseDown(object sender, MouseEventArgs e)
        {

            if (e.Button == System.Windows.Forms.MouseButtons.Left)
            {
                drawPointsFlag = true;
                mouseDrawpointsList = new List<PointF>();
            }
        }


        /// <summary>
        /// while moving points are stocked into array
        /// </summary>
        /// <param name="sender"></param>
        /// <param name="e"></param>
        private void pictureBox1_MouseMove(object sender, MouseEventArgs e)
        {
            if (drawPointsFlag)
            {
                if ((e.X < pictureBox1.Width) && (e.Y < pictureBox1.Height) && (e.Y > 0) && (e.X > 0))
                {
                    mouseDrawpointsList.Add(new PointF(e.X, e.Y));
                }

            }
        }

        /// <summary>
        /// all stocked points intialize the mouse draw points array , what will be used in the other functions
        /// </summary>
        /// <param name="sender"></param>
        /// <param name="e"></param>
        private void pictureBox1_MouseUp(object sender, MouseEventArgs e)
        {
            if (e.Button == System.Windows.Forms.MouseButtons.Left)
            {
                drawPointsFlag = false;
                if (DisplayDrawPoints != null)
                {
                    mouseDrawPoints = DisplayDrawPoints.ToList();
                    if (mouseDrawpointsList.Count > 3)
                        mouseDrawPoints.Add(mouseDrawpointsList.ToArray());

                    DisplayDrawPoints = mouseDrawPoints.ToArray();
                }

                if (mouseDrawpointsList.Count > 3)
                {

                    mouseDrawPoints.Add(mouseDrawpointsList.ToArray());
                    DisplayDrawPoints = mouseDrawPoints.ToArray();
                }
            }

        }


        #endregion

        #region Interface Interaction

        private void checkBoxShowVertices_CheckedChanged(object sender, EventArgs e)
        {
            IsShowVertices = checkBoxShowVertices.Checked;
        }


        private void trackBarSmoothingFactor_Scroll(object sender, EventArgs e)
        {
            SmoothingLimit = trackBarSmoothingFactor.Value;
            labelSmoothingFactor.Text = SmoothingLimit.ToString();
        }


        private void trackBarAttractionFactor_Scroll(object sender, EventArgs e)
        {
            AttractionFactor = trackBarAttractionFactor.Value;
            labelAttractionFactor.Text = AttractionFactor.ToString();
        }

        private void numericUpDownKernelSize_ValueChanged(object sender, EventArgs e)
        {
            int n = (int)numericUpDownKernelSize.Value;
            if (n % 2 == 0) { numericUpDownKernelSize.Value += 1; }

            KernelSize = (int)numericUpDownKernelSize.Value;
            Kernel = ComputeKernelValues(KernelSize);
            ComputeKernelGrad(Kernel, out GradientKernelX, out GradientKernelY);

        }

        #endregion

        private void richTextBox1_LinkClicked(object sender, LinkClickedEventArgs e)
        {
            System.Diagnostics.Process.Start(e.LinkText);
        }

       


    }
}


